#include "config.h"
#undef NULL
#include "lisp.h"
#include "termchar.h"
#include "dispextern.h"

#include <stdio.h>
#include <string.h>

#define min(x,y) ((x) > (y) ? (y) : (x))
#define max(x,y) ((x) < (y) ? (y) : (x))

#undef LONGBITS

#include <exec/types.h>
#include <intuition/intuition.h>
#include <devices/conunit.h>
#include <devices/inputevent.h>
#include <devices/timer.h>
#include <graphics/gfxbase.h>
#include <graphics/gfxmacros.h>
#include <utility/hooks.h>
#include <workbench/startup.h>
#include <workbench/workbench.h>

#include <proto/exec.h>
#include <proto/dos.h>
#include <proto/intuition.h>
#include <proto/graphics.h>
#include <proto/console.h>
#include <proto/timer.h>
#include <proto/diskfont.h>
#include <proto/wb.h>

extern struct Library *TimerBase, *WorkbenchBase;

#include "amiga.h"

#define SHIFT_MASK (IEQUALIFIER_LSHIFT | IEQUALIFIER_RSHIFT)
#define CONTROL_MASK IEQUALIFIER_CONTROL
#define META_MASK IEQUALIFIER_LALT

extern struct GfxBase *GfxBase;
extern struct IntuitionBase *IntuitionBase;
struct Library *DiskfontBase;

static struct Device *ConsoleDevice;
struct Window *emacs_win;
struct Screen *emacs_screen;
char *emacs_screen_name;
static struct TextFont *font;
int font_opened;
/* The reset string resets the console, turns off scrolling and sets up
   the foreground & background colors. */
#define CONSOLE_RESET "\x1b""c\x9b>1l\x9b""3%d;4%d;>%dm"
char reset_string[20]; /* Must be big enough for
			  printf(CONSOLE_RESET, foreground, background, background);
			  (0 <= foreground, background <= 7) */
int foreground = 1, background = 0;

static struct IOStdReq *emacs_console;

static struct NewWindow far emacs_window =
{
    0, 0,			    /* LeftEdge, TopEdge   */
    640, 200,			    /* Width, Height       */
    -1, -1,			    /* DetailPen, BlockPen */
    IDCMP_CLOSEWINDOW | IDCMP_RAWKEY | IDCMP_MOUSEBUTTONS | IDCMP_NEWSIZE |
    IDCMP_MENUPICK | IDCMP_MENUHELP,
    WFLG_CLOSEGADGET | WFLG_SIZEGADGET | WFLG_DRAGBAR | WFLG_DEPTHGADGET |
    WFLG_ACTIVATE | WFLG_SIMPLE_REFRESH,
    0,				    /* First Gadget        */
    0,				    /* CheckMark           */
    (UBYTE *)"GNU Emacs 18.58, Amiga port "VERS,
    0,				    /* Screen              */
    0,				    /* BitMap              */
    0, 0,			    /* Min Width, Height   */
    STDSCREENWIDTH, STDSCREENHEIGHT, /* Max Width, Height   */
    WBENCHSCREEN,		    /* Type */
};

#define emacs_icon_width 57
#define emacs_icon_height 55
#define emacs_icon_num_planes 1
#define emacs_icon_words_per_plane 220

UWORD chip emacs_icon_data[1][55][4] = {
    {
        0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
        0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
        0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0fe0,0x6000,
        0x0000,0x0000,0x0060,0x6000,0x0000,0x0000,0x0fff,0xe000,
        0x0000,0x0000,0x1800,0x2000,0x0000,0x0000,0x13ff,0xa000,
        0x0000,0x0000,0x1400,0xa000,0x0000,0x0000,0x3600,0xa000,
        0x0000,0x0000,0x0000,0xa000,0x0000,0x0000,0x0c00,0xa000,
        0x0000,0x0000,0x1e00,0xa000,0x0000,0x0000,0x0c00,0xa000,
        0x0000,0x0000,0x0000,0xa000,0x0000,0x0000,0x2100,0xa000,
        0x0000,0x0000,0x3300,0xa000,0x0000,0x0000,0x0c00,0xa000,
        0x003f,0xffff,0xffff,0xb000,0x001f,0xffff,0xffff,0x8000,
        0x004e,0x0000,0x0001,0xf000,0x00c6,0x00f0,0x0001,0x8000,
        0x00c6,0x0100,0x0001,0x8000,0x0006,0x0103,0x9201,0x8000,
        0x0006,0x013a,0x5201,0x8000,0x00c6,0x010a,0x5201,0x8000,
        0x00c6,0x010a,0x5601,0x8000,0x0086,0x00f2,0x4a01,0x8000,
        0x0006,0x0000,0x0001,0x8000,0x0046,0x0000,0x0001,0x8000,
        0x00c6,0x7c00,0x0001,0x8000,0x00c6,0x4000,0x0001,0x8000,
        0x0006,0x41d8,0xc319,0x8000,0x0006,0x7925,0x24a1,0x8000,
        0x00c6,0x4125,0x2419,0x8000,0x01c6,0x4125,0x2485,0x8000,
        0x0086,0x7d24,0xd319,0x8000,0x0007,0x0000,0x0003,0x8000,
        0x0003,0xffe3,0xffff,0x0000,0x0081,0xfff7,0xfffe,0x0000,
        0x01c0,0x0036,0x0000,0x0000,0x0180,0x0014,0x0f80,0x0000,
        0x0000,0x0014,0x1040,0x0000,0x0000,0x0014,0x2720,0x0000,
        0x0000,0x0012,0x28a0,0x0000,0x0080,0x000a,0x48a0,0x0000,
        0x01c0,0x0009,0x90a0,0x0000,0x0180,0x0004,0x20a0,0x0000,
        0x0000,0x0003,0xc0a0,0x0000,0x0000,0x0000,0x00a0,0x0000,
        0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
        0x0000,0x0000,0x0000,0x0000
    },
};

struct Image far emacs_icon = {
  0, 0,
  emacs_icon_width, emacs_icon_height, emacs_icon_num_planes,
  emacs_icon_data,
  3, 0,
  0
};
  
static struct Hook background_hook;

#define EVENTSIZE 32

static struct event {
    ULONG class;
    UWORD code, qual;
    WORD x, y;
} events[EVENTSIZE];
static int event_num, event_in, event_out;

Lisp_Object Vamiga_mouse_pos;
Lisp_Object Vamiga_mouse_item;
extern Lisp_Object MouseMap;
int amiga_remap_bsdel;
int amiga_mouse_initialized;

static int amiga_pos_x(int x)
{
    return (x - emacs_win->BorderLeft) / emacs_win->RPort->Font->tf_XSize;
}

static int amiga_pos_y(int y)
{
    return (y - emacs_win->BorderTop) / emacs_win->RPort->Font->tf_YSize;
}

static void amiga_change_size(void)
{
    int new_height = amiga_pos_y(emacs_win->Height - emacs_win->BorderBottom);
    int new_width = amiga_pos_x(emacs_win->Width - emacs_win->BorderRight);

    /* Hack to force redisplay */
    if (screen_height == new_height) screen_height--;
    change_screen_size(new_height, new_width, 0, !sleeping && !waiting_for_input, 1);
}

/* Get terminal size from system.
   Store number of lines into *heightp and width into *widthp.
   If zero or a negative number is stored, the value is not valid.  */

void get_window_size (widthp, heightp)
     int *widthp, *heightp;
{
    *heightp = amiga_pos_y(emacs_win->Height - emacs_win->BorderBottom);
    *widthp = amiga_pos_x(emacs_win->Width - emacs_win->BorderRight);
}

static int set_min_size(struct Window *win, struct TextFont *font,
			WORD *minw, WORD *minh)
{
    *minw = 11 * font->tf_XSize + win->BorderLeft + win->BorderRight;
    *minh = 4 * font->tf_YSize + win->BorderTop + win->BorderBottom;

    return (int)WindowLimits(emacs_win, *minw, *minh, 0, 0);
}

struct fill
{
    struct Layer *layer;
    struct Rectangle bounds;
    WORD offsetx, offsety;
};

static ULONG __asm __saveds fill_background(register __a2 struct RastPort *obj,
                                            register __a1 struct fill *msg)
{
    struct Layer *l;

    SetAPen(obj, background);
    SetDrMd(obj, JAM1);
    SetAfPt(obj, 0, 0);
    SetWrMsk(obj, 0xff);
    /* Gross hack starts here */
    l = obj->Layer;
    obj->Layer = 0;
    /* Stops */
    RectFill(obj, msg->bounds.MinX, msg->bounds.MinY,
	     msg->bounds.MaxX, msg->bounds.MaxY);
    /* Starts again */
    obj->Layer = l;
    /* And finally dies */

    return 0;
}

static void clear_window(void)
{
    SetAPen(emacs_win->RPort, background);
    RectFill(emacs_win->RPort, emacs_win->BorderLeft, emacs_win->BorderTop,
	     emacs_win->Width - emacs_win->BorderRight - 1,
	     emacs_win->Height - emacs_win->BorderBottom - 1);
}

static int make_reset_string(void)
{
    sprintf(reset_string, CONSOLE_RESET, foreground, background, background);
}

/* returns:
 *	-2 if msg is not class RAWKEY
 *	same as RawKeyConvert otherwise:
 *	buffer length if <= kbsize
 *	-1 else
 */
static DeadKeyConvert(struct IntuiMessage *msg, UBYTE *kbuffer, int kbsize,
		      struct KeyMap *kmap)
{
    static struct InputEvent ievent = {0, IECLASS_RAWKEY, 0, 0, 0};

    if (msg->Class != RAWKEY) return (-2);

    /* Do some keymapping ourselves to make emacs users happy */

    /* Ctrl-space becomes Ctrl-@ */
    if (msg->Code == 0x40 && msg->Qualifier & CONTROL_MASK)
    {
	*kbuffer = 0;
	return 1;
    }
    /* Backspace becomes DEL */
    if (msg->Code == 0x41 && amiga_remap_bsdel)
    {
	*kbuffer = 0177;
	return 1;
    }
    /* And DEL becomes CTRL-D */
    if (msg->Code == 0x46 && amiga_remap_bsdel)
    {
	*kbuffer = 04;
	return 1;
    }

    /* pack input event */
    ievent.ie_Code = msg->Code;

    /* Ignore meta in decoding keys */
    ievent.ie_Qualifier = msg->Qualifier & ~META_MASK;

    /* get previous codes from location pointed to by IAddress
     *  this pointer is valid until IntuiMessage is replied.
     */
    ievent.ie_position.ie_addr = *((APTR *)msg->IAddress);
    ievent.ie_position.ie_dead.ie_prev1DownQual &= ~META_MASK;
    ievent.ie_position.ie_dead.ie_prev2DownQual &= ~META_MASK;

    return ( RawKeyConvert(&ievent, kbuffer, kbsize, kmap) );
}

void check_window(int force)
{
    ULONG class;
    USHORT code;
    USHORT qualifier;
    UWORD mx, my;
    unsigned char buf[32];
    int buflen;
    struct IntuiMessage *msg;
    int mouse_event = FALSE;

    while (msg = (struct IntuiMessage *)GetMsg(emacs_win->UserPort))
    {
	class = msg->Class;
	code = msg->Code;
	qualifier = msg->Qualifier;
	mx = msg->MouseX; my = msg->MouseY;
	buflen = DeadKeyConvert(msg, buf, 32, 0);
	ReplyMsg(msg);

	switch (class)
	{
	    case IDCMP_CLOSEWINDOW: {
		enque(030, FALSE); enque(03, FALSE); /* ^X^C */
		break;
	    }
	    case IDCMP_RAWKEY: {
		if (buflen > 0)
		{
		    unsigned char *sbuf = buf;
		    int meta = qualifier & META_MASK;

		    /* Don't set META on CSI */
		    do {
			enque(*sbuf, meta && *sbuf != 0233);
			sbuf++;
		    } while (--buflen);
		}
		break;
	    }
	    case IDCMP_NEWSIZE: amiga_change_size(); break;
	    case IDCMP_MENUPICK: case IDCMP_MENUHELP:
	        if (code == MENUNULL) break; /* else fall through */
	    case IDCMP_MOUSEBUTTONS: {
		mouse_event = TRUE;
		if (event_num == EVENTSIZE) break;

		events[event_in].class = class;
		events[event_in].code = code;
		events[event_in].qual = qualifier;
		events[event_in].x = mx;
		events[event_in].y = my;
		event_num++;
		event_in = (event_in + 1) % EVENTSIZE;

		break;
	    }
	}
    }
    if (amiga_mouse_initialized && (force && event_num > 0 || mouse_event))
    {
	enque(AMIGASEQ, FALSE); enque('M', FALSE);
    }
}

DEFUN ("amiga-mouse-events", Famiga_mouse_events, Samiga_mouse_events, 0, 0, 0,
  "Return number of pending mouse events from Intuition.")
  ()
{
    register Lisp_Object tem;

    check_intuition ();

    XSET (tem, Lisp_Int, event_num);

    return tem;
}

DEFUN ("amiga-proc-mouse-event", Famiga_proc_mouse_event, Samiga_proc_mouse_event,
  0, 0, 0,
  "Pulls a mouse event out of the mouse event buffer and dispatches\n\
the appropriate function to act upon this event.")
  ()
{
    register Lisp_Object mouse_cmd;
    register char com_letter;
    register char key_mask;
    register Lisp_Object tempx;
    register Lisp_Object tempy;
    extern Lisp_Object get_keyelt ();
    extern int meta_prefix_char;
    struct event *ev;

    check_intuition ();

    if (event_num) {
	ev = &events[event_out];
	event_out = (event_out + 1) % EVENTSIZE;
	event_num--;
	if (ev->class == MOUSEBUTTONS)
	{
	    switch (ev->code)
	    {
		case SELECTDOWN: com_letter = 2; break;
		case SELECTUP: com_letter = 6; break;
		case MIDDLEDOWN: com_letter = 1; break;
		case MIDDLEUP: com_letter = 5; break;
		case MENUDOWN: com_letter = 0; break;
		case MENUUP: com_letter = 4; break;
		default: com_letter = 3; break;
	    }
	    XSET (tempx, Lisp_Int,
		  min (screen_width-1,
		       max (0, amiga_pos_x(ev->x))));
	    XSET (tempy, Lisp_Int,
		  min (screen_height-1,
		       max (0, amiga_pos_y(ev->y))));
	}
	else
	{
	    /* Must be Menu Pick or Help */
	    com_letter = ev->class == IDCMP_MENUPICK ? 3 : 7;

	    /* The parameters passed describe the selected item */
	    XSET (tempx, Lisp_Int, MENUNUM(ev->code));
	    XSET (tempy, Lisp_Int, ITEMNUM(ev->code));
	}
	if (ev->qual & META_MASK) com_letter |= 0x20;
	if (ev->qual & SHIFT_MASK) com_letter |= 0x10;
	if (ev->qual & CONTROL_MASK) com_letter |= 0x40;

	Vamiga_mouse_pos = Fcons (tempx, Fcons (tempy, Qnil));
	Vamiga_mouse_item = make_number (com_letter);
	mouse_cmd = get_keyelt (access_keymap (MouseMap, com_letter));
	if (NULL (mouse_cmd)) {
	    bell ();
	    Vamiga_mouse_pos = Qnil;
	}
	else return call1 (mouse_cmd, Vamiga_mouse_pos);
    }
    return Qnil;
}

DEFUN ("amiga-get-mouse-event", Famiga_get_mouse_event, Samiga_get_mouse_event,
  1, 1, 0,
  "Get next mouse event out of mouse event buffer (com-letter (x y)).\n\
ARG non-nil means return nil immediately if no pending event;\n\
otherwise, wait for an event.")
  (arg)
     Lisp_Object arg;
{
    register char com_letter;
    register char key_mask;

    register Lisp_Object tempx;
    register Lisp_Object tempy;
    struct event *ev;

    check_intuition ();

    if (NULL (arg))
    {
	amiga_consume_input(FALSE);
	while (!event_num)
	{
	    amiga_wait_for_input();
	    amiga_consume_input(FALSE);
	}
    }
    /*** ??? Surely you don't mean to busy wait??? */

    if (event_num) {
	ev = &events[event_out];
	event_out = (event_out + 1) % EVENTSIZE;
	event_num--;
	switch (ev->code)
	{
	    case SELECTDOWN: com_letter = 2; break;
	    case SELECTUP: com_letter = 6; break;
	    case MIDDLEDOWN: com_letter = 1; break;
	    case MIDDLEUP: com_letter = 5; break;
	    case MENUDOWN: com_letter = 0; break;
	    case MENUUP: com_letter = 4; break;
	    default: com_letter = 3; break;
	}
	if (ev->qual & META_MASK) com_letter |= 0x20;
	if (ev->qual & SHIFT_MASK) com_letter |= 0x10;
	if (ev->qual & CONTROL_MASK) com_letter |= 0x40;

	XSET (tempx, Lisp_Int,
	      min (screen_width-1,
		   max (0, amiga_pos_x(ev->x))));
	XSET (tempy, Lisp_Int,
	      min (screen_height-1,
		   max (0, amiga_pos_y(ev->y))));

	Vamiga_mouse_pos = Fcons (tempx, Fcons (tempy, Qnil));
	Vamiga_mouse_item = make_number (com_letter);
	return Fcons (com_letter, Fcons (Vamiga_mouse_pos, Qnil));
    }
    return Qnil;
}

DEFUN("amiga-set-foreground-color", Famiga_set_foreground_color,
  Samiga_set_foreground_color, 1, 1, "nPen number: ",
  "Use PEN as foreground color")
  (pen)
{
    int fg;

    CHECK_NUMBER(pen, 0);

    fg = XUINT (pen);
    if (pen > 7) error("Pen colors must be between 0 & 7");
    foreground = fg;
    make_reset_string();
    screen_puts (reset_string, strlen(reset_string));
    clear_window();
    amiga_change_size ();
    return Qnil;
}

DEFUN("amiga-set-background-color", Famiga_set_background_color,
  Samiga_set_background_color, 1, 1, "nPen number: ",
  "Use PEN as background color")
  (pen)
{
    int bg;

    CHECK_NUMBER(pen, 0);

    bg = XUINT (pen);
    if (pen > 7) error("Pen colors must be between 0 & 7");
    background = bg;
    make_reset_string();
    screen_puts (reset_string, strlen(reset_string));
    clear_window();
    amiga_change_size ();
    return Qnil;
}

DEFUN("amiga-set-font", Famiga_set_font, Samiga_set_font, 2, 2,
  "sFont: \n\
nSize: ",
  "Set font used for window to FONT with given HEIGHT.\n\
The font used must be non-proportional.")
  (wfont, height)
{
    struct TextAttr attr;
    struct TextFont *newfont;
    char *fname;
    struct Lisp_String *fstr;
    WORD minw, minh, oldmw, oldmh;

    CHECK_STRING (wfont, 0);
    CHECK_NUMBER (height, 0);

    check_intuition();

    fstr = XSTRING (wfont);
    fname = (char *)alloca (fstr->size + 6);
    strcpy (fname, fstr->data);
    strcat (fname, ".font");
    attr.ta_Name = fname;
    attr.ta_YSize = XFASTINT (height);
    attr.ta_Style = 0;
    attr.ta_Flags = 0;
    newfont = OpenDiskFont (&attr);

    if (!newfont)
	error ("Font %s %d not found", fstr->data, XFASTINT (height));
    if (newfont->tf_Flags & FPF_PROPORTIONAL)
    {
	CloseFont(newfont);
	error ("Font %s %d is proportional", fstr->data, XFASTINT (height));
    }

    if (!set_min_size(emacs_win, newfont, &minw, &minh))
    {
	CloseFont(newfont);
	if (!set_min_size(emacs_win, font, &oldmw, &oldmh))
	{
	    fprintf(stderr, "Failed to restore old minimum window size, aborting\n");
	    exit(1);
	}
	error("Window is too small for this font, need at least %d(w) by %d(h)",
	      minw, minh);
    }

    if (font_opened) CloseFont(font);
    font_opened = TRUE;

    SetFont (emacs_win->RPort, newfont);
    font = newfont;
    /* Reset console to use new font */
    screen_puts (reset_string, strlen(reset_string));
    clear_window();
    amiga_change_size ();
    return Qnil;
}

DEFUN("amiga-set-geometry", Famiga_set_geometry, Samiga_set_geometry, 4, 5, 0,
  "Set Emacs window geometry and screen.\n\
First 4 parameters are the (X,Y) position of the top-left corner of the window\n\
and its WIDTH and HEIGHT. These must be big enough for an 11x4 characters window.\n\
The optional argument SCREEN specifies which screen to use, nil stands for the\n\
same screen as the window is on, t stands for the default public screen (normally\n\
the Workbench), a string specifies a given public screen.")
  (x, y, w, h, scr)
Lisp_Object x, y, w, h, scr;
{
    struct Screen *newscreen;
    struct Window *newwindow;
    struct IOStdReq *newconsole;
    WORD minw, minh;

    CHECK_NUMBER(x, 0);
    CHECK_NUMBER(y, 0);
    CHECK_NUMBER(w, 0);
    CHECK_NUMBER(h, 0);

    if (NULL (scr)) newscreen = emacs_screen;
    else if (scr == Qt) newscreen = 0;
    else
    {
	CHECK_STRING(scr, 0);
	newscreen = LockPubScreen(XSTRING (scr)->data);
	if (!newscreen) error("Unknown public screen %s", XSTRING (scr)->data);
    }
    newwindow = OpenWindowTags(&emacs_window,
			       WA_Left, XUINT (x), WA_Top, XUINT (y),
			       WA_Width, XUINT (w), WA_Height, XUINT (h),
			       WA_PubScreen, newscreen,
       			       WA_BackFill, &background_hook,
			       WA_MenuHelp, TRUE,
			       TAG_END);
    UnlockPubScreen(0L, newscreen);
    if (!newwindow) error("Failed to open desired window");
    if (!set_min_size(newwindow, font, &minw, &minh))
    {
	CloseWindow(newwindow);
	error("Window is too small, need at least %d(w) by %d(h)", minw, minh);
    }
    SetFont(emacs_win->RPort, font);

    newconsole = (struct IOStdReq *)device_open("console.device", CONU_CHARMAP,
						CONFLAG_NODRAW_ON_NEWSIZE,
						(APTR)newwindow, sizeof(*newwindow),
						sizeof(struct IOStdReq));
    if (!newconsole)
    {
	CloseWindow(newwindow);
	error("Failed to open console for window (No memory?)");
    }
    suspend_menus();
    if (!set_menu_window(newwindow))
    {
	device_close(newconsole);
	CloseWindow(newwindow);
	resume_menus();
	error("Failed to setup menus (No memory?)");
    }
    /* Use new window */
    device_close(emacs_console);
    emacs_console = newconsole;
    inputsig &= ~(1L << emacs_win->UserPort->mp_SigBit);
    emacs_screen = newscreen;
    if (emacs_screen_name) free (emacs_screen_name);
    emacs_screen_name = (char *)xmalloc (XSTRING (scr)->size + 1);
    strcpy(emacs_screen_name, XSTRING (scr)->data);
    CloseWindow(emacs_win);
    emacs_win = newwindow;
    inputsig |= 1L << emacs_win->UserPort->mp_SigBit;
    resume_menus();
    /* Turn off scrolling */
    screen_puts (reset_string, strlen(reset_string));
    clear_window();
    amiga_change_size ();

    return Qnil;
}


void Iconify(struct Image *im)
{
    struct AppIcon *icon;
    struct MsgPort *port;
    static struct DiskObject wbicon = {
	0, 0,
        { 0 },
	0, 0, 0,
	NO_ICON_POSITION, NO_ICON_POSITION
	};


    if (port = CreateMsgPort())
    {
	wbicon.do_Gadget.Width = im->Width;
	wbicon.do_Gadget.Height = im->Height;
	wbicon.do_Gadget.GadgetRender = (APTR)im;

	if (icon = AddAppIconA(0, 0, "Emacs", port, 0, &wbicon, 0))
	{
	    struct AppMessage *msg;
	    int quit = FALSE;

	    while (!quit)
	    {
		WaitPort(port);
		while (msg = (struct AppMessage *)GetMsg(port))
		{
		    quit = TRUE;
		    ReplyMsg(msg);
		}
	    }
	    RemoveAppIcon(icon);
	}
    }
}

DEFUN("amiga-iconify", Famiga_iconify, Samiga_iconify, 0, 0, "",
  "Iconify emacs window. \n\
The routine returns when emacs is deiconified.")
  ()
{
    WORD minw, minh;
    UWORD x, y, w, h;

    /* Close window */
    suspend_menus();
    device_close(emacs_console);
    inputsig &= ~(1L << emacs_win->UserPort->mp_SigBit);
    x = emacs_win->LeftEdge; y = emacs_win->TopEdge;
    w = emacs_win->Width; h = emacs_win->Height;
    CloseWindow(emacs_win);

    Iconify(&emacs_icon);

    /* Reopen window */
    emacs_screen = LockPubScreen(emacs_screen_name);
    emacs_win = OpenWindowTags(&emacs_window,
			       WA_Left, x, WA_Top, y, WA_Width, w, WA_Height, h,
			       WA_PubScreen, emacs_screen,
			       WA_BackFill, &background_hook,
			       WA_MenuHelp, TRUE,
			       TAG_END);
    UnlockPubScreen(0L, emacs_screen);
    if (emacs_win)
    {
	SetFont(emacs_win->RPort, font);

	if (set_min_size(emacs_win, font, &minw, &minh) &&
	    set_menu_window(emacs_win) && 
	    (emacs_console = (struct IOStdReq *)
	     device_open("console.device", CONU_CHARMAP, CONFLAG_NODRAW_ON_NEWSIZE,
			 (APTR)emacs_win, sizeof(*emacs_win),
			 sizeof(struct IOStdReq))))
	{
	    inputsig |= 1L << emacs_win->UserPort->mp_SigBit;
	    resume_menus();
	    /* Turn off scrolling */
	    screen_puts (reset_string, strlen(reset_string));
	    clear_window();
	    amiga_change_size ();

	    return Qnil;
	}
	CloseWindow(emacs_win);
    }
    fprintf(stderr, "Iconify: failed to reinitialise (No memory?)\n");
    exit(1);
}

struct EClockVal scount[16], ecount[16];
long total[16], counting[16], nb[16], susp[16];

void start_count(int n)
{
/*    nb[n]++;
    if (counting[n]) printf("Restarted %d\n", n);
    counting[n] = 1;
    ReadEClock(&scount[n]);*/
}

void stop_count(int n)
{
    if (counting[n])
    {
	ReadEClock(&ecount[n]);
	counting[n] = 0;

	total[n] += ecount[n].ev_lo - scount[n].ev_lo;
    }
}

void suspend_count(int n)
{
    if (counting[n] && susp[n]++ == 0)
    {
	ReadEClock(&ecount[n]);
	total[n] += ecount[n].ev_lo - scount[n].ev_lo;
    }
}

void resume_count(int n)
{
    if (counting[n] && --susp[n] == 0) ReadEClock(&scount[n]);
}

void screen_puts(char *str, unsigned int len)
{
    int i;

    emacs_console->io_Command = CMD_WRITE;
    emacs_console->io_Data    = (APTR)str;
    emacs_console->io_Length  = len;

/*    start_count(0);
    for (i = 1; i <= 6; i++) suspend_count(i);*/
    DoIO(emacs_console);
/*    for (i = 1; i <= 6; i++) resume_count(i);
    stop_count(0);*/
}

void syms_of_amiga_screen(void)
{
    DEFVAR_LISP ("amiga-mouse-item", &Vamiga_mouse_item,
		 "Encoded representation of last mouse click, corresponding to\n\
numerical entries in amiga-mouse-map.");
    Vamiga_mouse_item = Qnil;
    DEFVAR_LISP ("amiga-mouse-pos", &Vamiga_mouse_pos,
		 "Current x-y position of mouse by row, column as specified by font.");
    Vamiga_mouse_pos = Qnil;

    DEFVAR_BOOL ("amiga-remap-bsdel", &amiga_remap_bsdel,
"*If true, map DEL to Ctrl-D and Backspace to DEL. \n\
This is the most convenient (and default) setting. If nil, don't remap.");
    amiga_remap_bsdel = 1;

    DEFVAR_BOOL ("amiga-mouse-initialized", &amiga_mouse_initialized,
"Set to true once lisp has been setup to process mouse commands.\n\
No mouse processing request (C-\ M) will be queued while this is nil.");
    amiga_mouse_initialized = 0;

    defsubr (&Samiga_mouse_events);
    defsubr (&Samiga_proc_mouse_event);
    defsubr (&Samiga_get_mouse_event);
    defsubr (&Samiga_set_font);
    defsubr (&Samiga_set_geometry);
    defsubr (&Samiga_set_background_color);
    defsubr (&Samiga_set_foreground_color);
    defsubr (&Samiga_iconify);
}

int init_amiga_screen(void)
{
    int minw, minh;

    event_num = event_in = event_out = 0;

    if (!((IntuitionBase = (struct IntuitionBase *)
	   OpenLibrary("intuition.library", 37L)) &&
	  (GfxBase = (struct GfxBase *)OpenLibrary("graphics.library", 0L)) &&
	  (DiskfontBase = OpenLibrary("diskfont.library", 0L)) &&
	  (WorkbenchBase = OpenLibrary("workbench.library", 37))))
    {
	fprintf(stderr, "Need version 2.04 and diskfont.library!\n");
	return FALSE;
    }
    background_hook.h_Entry = fill_background;
    emacs_screen = 0; emacs_screen_name = 0;
    if ((emacs_win = OpenWindowTags(&emacs_window,
				    WA_BackFill, &background_hook,
				    WA_MenuHelp, TRUE,
				    TAG_END)))
    {
	WORD minw, minh;

	font = GfxBase->DefaultFont;
	SetFont(emacs_win->RPort, font);
	if (!set_min_size(emacs_win, font, &minw, &minh))
	{
	    fprintf(stderr, "Default font is too large for a 640 by 200 window\n");
	    CloseWindow(emacs_win);
	    return FALSE;
	}
	if (emacs_console =
	    (struct IOStdReq *)device_open("console.device", CONU_CHARMAP,
					   CONFLAG_NODRAW_ON_NEWSIZE,
					   (APTR)emacs_win, sizeof(*emacs_win),
					   sizeof(struct IOStdReq)))
	{
	    ConsoleDevice = emacs_console->io_Device;
	    inputsig |= 1L << emacs_win->UserPort->mp_SigBit;
	    make_reset_string();
	    screen_puts (reset_string, strlen(reset_string));
	    clear_window();
	}
	else
	{
	    fprintf(stderr, "Console didn't open!\n");
	    return FALSE;
	}
    }
    else
    {
	fprintf(stderr, "Window will not open!\n");
	return FALSE;
    }
    return init_amiga_menu();
}

void cleanup_amiga_screen(void)
{
    cleanup_amiga_menu();
    device_close(emacs_console);
    if (emacs_win) CloseWindow(emacs_win);
    if (font_opened) CloseFont(font);
    if (IntuitionBase) CloseLibrary(IntuitionBase);
    if (GfxBase) CloseLibrary(GfxBase);
    if (WorkbenchBase) CloseLibrary(WorkbenchBase);
}
